;################################################################################
;#										#
;# chipbasic - single chip basic computer with ATMega 16			#
;# EEPROM memory interface							#
;# copyright (c) 2006 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;#										#
;# This program is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU General Public License			#
;# as published by the Free Software Foundation; either version 2		#
;# of the License, or (at your option) any later version.			#
;#										#
;# This program is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

;------------------------------------------------------------------------------
; read one byte from program memory
;------------------------------------------------------------------------------
ee_read:	sbic	EECR,EEWE
		rjmp	ee_read
		out	EEARH,YH
		out	EEARL,YL
		sbi	EECR,EERE
		in	tempreg1,EEDR
		adiw	YL,1
		ret

;------------------------------------------------------------------------------
; read line from program memory to line buffer
;------------------------------------------------------------------------------
ee_readline:	mov	XL,r12			;linenumber
		lds	XH,bas_ram		;cached line
		cp	XH,XL			;is equal?
		breq	ee_readline2		;yes
		sts	bas_ram,XL		;store as last
		mov	ctrl,XL
;read line XL
ee_readline0:	ldi	ZL,LOW(bas_linebuf)	;buffer address
		ldi	ZH,HIGH(bas_linebuf)
		ldi	XH,mem_linelen		;bytes/line
		mul	XH,ctrl
		mov	YL,r0
		mov	YH,r1
ee_readline1:	rcall	ee_read			;get byte
		st	Z+,tempreg1
		cpi	tempreg1,0xff		;EOL
		breq	ee_readline2
		dec	XH
		brne	ee_readline1
ee_readline2:	ldi	tempreg1,0xff		;an extra EOL
		st	Z+,tempreg1
		ret					

;------------------------------------------------------------------------------
; read name from program memory to Z+
;------------------------------------------------------------------------------
ee_readname:	ldi	YL,LOW(500)
		ldi	YH,HIGH(500)
		ldi	XH,10			;bytes/name
ee_readname1:	rcall	ee_read			;get byte
		st	Z+,tempreg1
		dec	XH
		brne	ee_readname1
		ret					
		
;------------------------------------------------------------------------------
; write one byte to program memory
;------------------------------------------------------------------------------
ee_write:	sbic	EECR,EEWE
		rjmp	ee_write
		out	EEARH,YH
		out	EEARL,YL
		sbi	EECR,EERE
		in	tempreg5,EEDR
		cp	tempreg5,tempreg1
		brne	ee_write_1
		adiw	YL,1
		ret	
ee_write_1:	sbic	EECR,EEWE
		rjmp	ee_write_1
		out	EEDR,tempreg1
		sbi	EECR,EEMWE
		sbi	EECR,EEWE
ee_write_2:	sbic	EECR,EEWE		;verify
		rjmp	ee_write_2
		sbi	EECR,EERE
		in	tempreg5,EEDR
		cp	tempreg1,tempreg5		
		brne	ee_write		;repeat programming
		adiw	YL,1
		ret

;------------------------------------------------------------------------------
; write line [ctrl] from line buffer to program memory
;------------------------------------------------------------------------------
ee_writeline:	ldi	ZL,LOW(bas_linebuf)	;buffer address
		ldi	ZH,HIGH(bas_linebuf)
		ldi	XH,mem_linelen		;length of EE-line
		mul	XH,ctrl			;linenumber
		mov	YL,r0		
		mov	YH,r1
ee_writeline4:	ld	tempreg1,Z+
		rcall	ee_write
		cpi	tempreg1,0xff
		breq	ee_writel_16
		dec	XH
		brne	ee_writeline4		
ee_writel_16:	ret				;exit if full	

;------------------------------------------------------------------------------
; store name from Z+ to program memory
;------------------------------------------------------------------------------
ee_writename:	ldi	YL,LOW(500)
		ldi	YH,HIGH(500)
		ldi	XH,10			;bytes/name
ee_writename1:	ld	tempreg1,Z+
		rcall	ee_write		;write byte
		dec	XH
		brne	ee_writename1
		ret					
